#include "ncbind/ncbind.hpp"
#include <vector>
using namespace std;
#include <process.h>

#define WM_SAVE_TLG_PROGRESS (WM_APP+4)
#define WM_SAVE_TLG_DONE     (WM_APP+5)

iTJSDispatch2 *getLayerClass(void)
{
  tTJSVariant var;
  TVPExecuteExpression(TJS_W("Layer"), &var);
  return  var.AsObjectNoAddRef();
}

iTJSDispatch2 *getLayerAssignImages(void)
{
  tTJSVariant var;
  TVPExecuteExpression(TJS_W("Layer.assignImages"), &var);
  return  var.AsObjectNoAddRef();
}

#include "compress.hpp"
#include "savetlg5.hpp"
#include "savepng.hpp"

//---------------------------------------------------------------------------
// EChEg
//---------------------------------------------------------------------------

class WindowSaveImage;

/**
 * Z[uXbhp
 */
class SaveInfo {

	friend class WindowSaveImage;
	
protected:

	// ϐ
	WindowSaveImage *notify; //< ʒm
	tTJSVariant layer; //< C
	tTJSVariant filename; //< t@C
	tTJSVariant info;  //< ۑp^O
	bool canceled;        //< LZw
	tTJSVariant handler;  //< nhl
	tTJSVariant progressPercent; //< isx
	
protected:
	/**
	 * ݂̏Ԃ̒ʒm
	 * @param percent p[Zg
	 */
	bool progress(int percent);

	/**
	 * Ăяop
	 */
	static bool progressFunc(int percent, void *userData) {
		SaveInfo *self = (SaveInfo*)userData;
		return self->progress(percent);
	}
	
	// o߃CxgM
	void eventProgress(iTJSDispatch2 *objthis) {
		tTJSVariant *vars[] = {&handler, &progressPercent, &layer, &filename};
		objthis->FuncCall(0, L"onSaveLayerImageProgress", NULL, NULL, 4, vars, objthis);
	}

	// ICxgM
	void eventDone(iTJSDispatch2 *objthis) {
		tTJSVariant result = canceled ? 1 : 0;
		tTJSVariant *vars[] = {&handler, &result, &layer, &filename};
		objthis->FuncCall(0, L"onSaveLayerImageDone", NULL, NULL, 4, vars, objthis);
	}
	
public:
	// RXgN^
	SaveInfo(int handler, WindowSaveImage *notify, tTJSVariant layer, const tjs_char *filename, tTJSVariant info)
		: handler(handler), notify(notify), layer(layer), filename(filename), info(info), canceled(false) {}
	
	// fXgN^
	~SaveInfo() {}

	// nh擾
	int getHandler() {
		return (int)handler;
	}
	
 	// Jn
	void start();

	// LZ
	void cancel() {
		canceled = true;
	}

	// I
	void stop() {
		canceled = true;
		notify = NULL;
	}
};

/**
 * EChEɃCZ[u@\g
 */
class WindowSaveImage {

protected:
	iTJSDispatch2 *objthis; //< IuWFNg̎Q

	vector<SaveInfo*> saveinfos; //< Z[uێp

	// sXbh
	static void checkThread(void *data) {
		((SaveInfo*)data)->start();
	}

	// oߒʒm
	void eventProgress(SaveInfo *sender) {
		int handler = sender->getHandler();
		if (saveinfos[handler] == sender) {
			sender->eventProgress(objthis);
		}
	}

	// Iʒm
	void eventDone(SaveInfo *sender) {
		int handler = sender->getHandler();
		if (saveinfos[handler] == sender) {
			saveinfos[handler] = NULL;
			sender->eventDone(objthis);
		}
		delete sender;
	}

	/*
	 * EChECxgV[o
	 */
	static bool __stdcall receiver(void *userdata, tTVPWindowMessage *Message) {
		if (Message->Msg == WM_SAVE_TLG_PROGRESS) {
			iTJSDispatch2 *obj = (iTJSDispatch2*)userdata;
			WindowSaveImage *self = ncbInstanceAdaptor<WindowSaveImage>::GetNativeInstance(obj);
			if (self) {
				self->eventProgress((SaveInfo*)Message->WParam);
			}
			return true;
		} else if (Message->Msg == WM_SAVE_TLG_DONE) {
			iTJSDispatch2 *obj = (iTJSDispatch2*)userdata;
			WindowSaveImage *self = ncbInstanceAdaptor<WindowSaveImage>::GetNativeInstance(obj);
			if (self) {
				self->eventDone((SaveInfo*)Message->WParam);
			}
			return true;
		}
		return false;
	}

	// [UbZ[WV[o̓o^/
	void setReceiver(tTVPWindowMessageReceiver receiver, bool enable) {
		tTJSVariant mode     = enable ? (tTVInteger)(tjs_int)wrmRegister : (tTVInteger)(tjs_int)wrmUnregister;
		tTJSVariant proc     = (tTVInteger)(tjs_int)receiver;
		tTJSVariant userdata = (tTVInteger)(tjs_int)objthis;
		tTJSVariant *p[] = {&mode, &proc, &userdata};
		if (objthis->FuncCall(0, L"registerMessageReceiver", NULL, NULL, 4, p, objthis) != TJS_S_OK) {
			TVPThrowExceptionMessage(L"can't regist user message receiver");
		}
	}

public:

	/**
	 * RXgN^
	 */
	WindowSaveImage(iTJSDispatch2 *objthis) : objthis(objthis) {
		setReceiver(receiver, true);
	}

	/**
	 * fXgN^
	 */
	~WindowSaveImage() {
		setReceiver(receiver, false);
		for (int i=0;i<(int)saveinfos.size();i++) {
			SaveInfo *saveinfo = saveinfos[i];
			if (saveinfo) {
				saveinfo->stop();
				saveinfos[i] = NULL;
			}
		}
	}

	/**
	 * bZ[WM
	 * @param msg bZ[W
	 * @param wparam WPARAM
	 * @param lparam LPARAM
	 */
	void postMessage(UINT msg, WPARAM wparam=NULL, LPARAM lparam=NULL) {
		// EBhEnh擾Ēʒm
		tTJSVariant val;
		objthis->PropGet(0, TJS_W("HWND"), NULL, &val, objthis);
		HWND hwnd = reinterpret_cast<HWND>((tjs_int)(val));
		::PostMessage(hwnd, msg, wparam, lparam);
	}

	/**
	 * CZ[uJn
	 * @param layer C
	 * @param filename t@C
	 * @param info ^O
	 */
	int startSaveLayerImage(tTJSVariant layer, const tjs_char *filename, tTJSVariant info) {
		int handler = saveinfos.size();
		for (int i=0;i<(int)saveinfos.size();i++) {
			if (saveinfos[i] == NULL) {
				handler = i;
				break;
			}
		}
		if (handler >= (int)saveinfos.size()) {
			saveinfos.resize(handler + 1);
		}

		// ۑpɃC𕡐
		tTJSVariant newLayer;
		{
			// VC𐶐
			tTJSVariant window(objthis, objthis);
			tTJSVariant primaryLayer;
			objthis->PropGet(0, L"primaryLayer", NULL, &primaryLayer, objthis);
			tTJSVariant *vars[] = {&window, &primaryLayer};
			iTJSDispatch2 *obj;
			if (TJS_SUCCEEDED(getLayerClass()->CreateNew(0, NULL, NULL, &obj, 2, vars, objthis))) {

				// OÂ
				tTJSVariant name = "saveLayer:";
				name +=filename;
				obj->PropSet(0, L"name", NULL, &name, obj);

				// C̉摜𕡐
				tTJSVariant *param[] = {&layer};
				if (TJS_SUCCEEDED(getLayerAssignImages()->FuncCall(0, NULL, NULL, NULL, 1, param, obj))) {
					newLayer = tTJSVariant(obj, obj);
					obj->Release();
				} else {
					obj->Release();
					TVPThrowExceptionMessage(L"ۑpCւ̉摜̕Ɏs܂");
				}
			} else {
				TVPThrowExceptionMessage(L"ۑpC̐Ɏs܂");
			}
		}
		SaveInfo *saveInfo = new SaveInfo(handler, this, newLayer, filename, info);
		saveinfos[handler] = saveInfo;
		_beginthread(checkThread, 0, saveInfo);
		return handler;
	}
	
	/**
	 * CZ[ũLZ
	 */
	void cancelSaveLayerImage(int handler) {
		if (handler < (int)saveinfos.size() && saveinfos[handler] != NULL) {
			saveinfos[handler]->cancel();
		}
	}

	/**
	 * CZ[u̒~
	 */
	void stopSaveLayerImage(int handler) {
		if (handler < (int)saveinfos.size() && saveinfos[handler] != NULL) {
			saveinfos[handler]->stop();
			saveinfos[handler] = NULL;
		}
	}
};


/**
 * ݂̏Ԃ̒ʒm
 * @param percent p[Zg
 */
bool
SaveInfo::progress(int percent)
{
	if ((int)progressPercent != percent) {
		progressPercent = percent;
		if (notify) {
			notify->postMessage(WM_SAVE_TLG_PROGRESS, (WPARAM)this);
			Sleep(0);
		}
	}
	return canceled;
}

/*
 * ۑJn
 */
void
SaveInfo::start()
{
	iTJSDispatch2  *lay = layer.AsObjectNoAddRef();
	iTJSDispatch2  *nfo = info.Type() == tvtObject ? info.AsObjectNoAddRef() : NULL;
	const tjs_char *fn  = filename.GetString();
	ttstr ext(TVPExtractStorageExt(ttstr(fn)));
	ext.ToLowerCase();
	// 摜Z[uigqʁj
	if (ext == TJS_W(".png")) {
		CompressAndSave<CompressPNG >::saveLayerImage(lay, fn, nfo, progressFunc, (void*)this);
	} else {
		CompressAndSave<CompressTLG5>::saveLayerImage(lay, fn, nfo, progressFunc, (void*)this);
	}
	// ʒm
	if (notify) {
		notify->postMessage(WM_SAVE_TLG_DONE, (WPARAM)this);
		Sleep(0);
	} else {
		delete this;
	}
}

//---------------------------------------------------------------------------

// CX^XQb^
NCB_GET_INSTANCE_HOOK(WindowSaveImage)
{
	NCB_INSTANCE_GETTER(objthis) { // objthis  iTJSDispatch2* ^̈Ƃ
		ClassT* obj = GetNativeInstance(objthis);	// lCeBuCX^X|C^擾
		if (!obj) {
			obj = new ClassT(objthis);				// Ȃꍇ͐
			SetNativeInstance(objthis, obj);		// objthis  obj lCeBuCX^XƂēo^
		}
		return obj;
	}
};

NCB_ATTACH_CLASS_WITH_HOOK(WindowSaveImage, Window) {
	NCB_METHOD(startSaveLayerImage);
	NCB_METHOD(cancelSaveLayerImage);
	NCB_METHOD(stopSaveLayerImage);
};

